from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.api.featureservice import feature_service_request__client_pb2 as _feature_service_request__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Text, Union

CACHED_MISSING_DATA: Status
CACHED_PRESENT: Status
CACHED_UNKNOWN: Status
DEFAULT: FeatureServiceType
DESCRIPTOR: _descriptor.FileDescriptor
DYNAMODB_RESPONSE_SIZE_LIMIT_EXCEEDED: SloIneligibilityReason
MISSING_DATA: Status
PRESENT: Status
REDIS_LATENCY_LIMIT_EXCEEDED: SloIneligibilityReason
REDIS_RESPONSE_SIZE_LIMIT_EXCEEDED: SloIneligibilityReason
UNKNOWN: SloIneligibilityReason
UNKNOWN_STATUS: Status
WILDCARD: FeatureServiceType
array: FeatureServerDataType
boolean: FeatureServerDataType
float32: FeatureServerDataType
float64: FeatureServerDataType
int64: FeatureServerDataType
map: FeatureServerDataType
missing_type: FeatureServerDataType
string: FeatureServerDataType
string_array: FeatureServerDataType
struct: FeatureServerDataType
timestamp: FeatureServerDataType

class BatchMetadata(_message.Message):
    __slots__ = ["batch_slo_info", "features", "join_keys", "slo_info"]
    class Item(_message.Message):
        __slots__ = ["data_type", "description", "effective_time", "name", "status", "tags"]
        class TagsEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: ClassVar[int]
            VALUE_FIELD_NUMBER: ClassVar[int]
            key: str
            value: str
            def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
        DATA_TYPE_FIELD_NUMBER: ClassVar[int]
        DESCRIPTION_FIELD_NUMBER: ClassVar[int]
        EFFECTIVE_TIME_FIELD_NUMBER: ClassVar[int]
        NAME_FIELD_NUMBER: ClassVar[int]
        STATUS_FIELD_NUMBER: ClassVar[int]
        TAGS_FIELD_NUMBER: ClassVar[int]
        data_type: FeatureServerComplexDataType
        description: str
        effective_time: _timestamp_pb2.Timestamp
        name: str
        status: _containers.RepeatedScalarFieldContainer[Status]
        tags: _containers.ScalarMap[str, str]
        def __init__(self, name: Optional[str] = ..., effective_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., data_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ..., status: Optional[Iterable[Union[Status, str]]] = ..., description: Optional[str] = ..., tags: Optional[Mapping[str, str]] = ...) -> None: ...
    BATCH_SLO_INFO_FIELD_NUMBER: ClassVar[int]
    FEATURES_FIELD_NUMBER: ClassVar[int]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    SLO_INFO_FIELD_NUMBER: ClassVar[int]
    batch_slo_info: BatchSloInfo
    features: _containers.RepeatedCompositeFieldContainer[BatchMetadata.Item]
    join_keys: _containers.RepeatedCompositeFieldContainer[BatchMetadata.Item]
    slo_info: _containers.RepeatedCompositeFieldContainer[SloInfo]
    def __init__(self, features: Optional[Iterable[Union[BatchMetadata.Item, Mapping]]] = ..., join_keys: Optional[Iterable[Union[BatchMetadata.Item, Mapping]]] = ..., slo_info: Optional[Iterable[Union[SloInfo, Mapping]]] = ..., batch_slo_info: Optional[Union[BatchSloInfo, Mapping]] = ...) -> None: ...

class BatchSloInfo(_message.Message):
    __slots__ = ["server_time_seconds", "slo_eligible", "slo_ineligibility_reasons", "slo_server_time_seconds", "store_max_latency"]
    SERVER_TIME_SECONDS_FIELD_NUMBER: ClassVar[int]
    SLO_ELIGIBLE_FIELD_NUMBER: ClassVar[int]
    SLO_INELIGIBILITY_REASONS_FIELD_NUMBER: ClassVar[int]
    SLO_SERVER_TIME_SECONDS_FIELD_NUMBER: ClassVar[int]
    STORE_MAX_LATENCY_FIELD_NUMBER: ClassVar[int]
    server_time_seconds: float
    slo_eligible: bool
    slo_ineligibility_reasons: _containers.RepeatedScalarFieldContainer[SloIneligibilityReason]
    slo_server_time_seconds: float
    store_max_latency: float
    def __init__(self, slo_eligible: bool = ..., slo_server_time_seconds: Optional[float] = ..., slo_ineligibility_reasons: Optional[Iterable[Union[SloIneligibilityReason, str]]] = ..., server_time_seconds: Optional[float] = ..., store_max_latency: Optional[float] = ...) -> None: ...

class FeatureServerComplexDataType(_message.Message):
    __slots__ = ["element_type", "fields", "key_type", "type", "value_type"]
    ELEMENT_TYPE_FIELD_NUMBER: ClassVar[int]
    FIELDS_FIELD_NUMBER: ClassVar[int]
    KEY_TYPE_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    VALUE_TYPE_FIELD_NUMBER: ClassVar[int]
    element_type: FeatureServerComplexDataType
    fields: _containers.RepeatedCompositeFieldContainer[FeatureServerStructField]
    key_type: FeatureServerComplexDataType
    type: FeatureServerDataType
    value_type: FeatureServerComplexDataType
    def __init__(self, type: Optional[Union[FeatureServerDataType, str]] = ..., element_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ..., fields: Optional[Iterable[Union[FeatureServerStructField, Mapping]]] = ..., key_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ..., value_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ...) -> None: ...

class FeatureServerStructField(_message.Message):
    __slots__ = ["data_type", "name"]
    DATA_TYPE_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    data_type: FeatureServerComplexDataType
    name: str
    def __init__(self, name: Optional[str] = ..., data_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ...) -> None: ...

class FeatureServiceLocator(_message.Message):
    __slots__ = ["feature_package_id", "feature_package_name", "feature_service_id", "feature_service_name", "feature_view_id", "feature_view_name", "workspace_name"]
    FEATURE_PACKAGE_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_PACKAGE_NAME_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_NAME_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_NAME_FIELD_NUMBER: ClassVar[int]
    feature_package_id: str
    feature_package_name: str
    feature_service_id: str
    feature_service_name: str
    feature_view_id: str
    feature_view_name: str
    workspace_name: str
    def __init__(self, feature_service_id: Optional[str] = ..., feature_service_name: Optional[str] = ..., feature_package_id: Optional[str] = ..., feature_package_name: Optional[str] = ..., feature_view_id: Optional[str] = ..., feature_view_name: Optional[str] = ..., workspace_name: Optional[str] = ...) -> None: ...

class GetFeatureServiceStateRequest(_message.Message):
    __slots__ = ["feature_service_locator"]
    FEATURE_SERVICE_LOCATOR_FIELD_NUMBER: ClassVar[int]
    feature_service_locator: FeatureServiceLocator
    def __init__(self, feature_service_locator: Optional[Union[FeatureServiceLocator, Mapping]] = ...) -> None: ...

class GetFeatureServiceStateResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class GetFeaturesBatchResponse(_message.Message):
    __slots__ = ["metadata", "result"]
    METADATA_FIELD_NUMBER: ClassVar[int]
    RESULT_FIELD_NUMBER: ClassVar[int]
    metadata: BatchMetadata
    result: _containers.RepeatedCompositeFieldContainer[GetFeaturesResult]
    def __init__(self, result: Optional[Iterable[Union[GetFeaturesResult, Mapping]]] = ..., metadata: Optional[Union[BatchMetadata, Mapping]] = ...) -> None: ...

class GetFeaturesResponse(_message.Message):
    __slots__ = ["metadata", "result"]
    METADATA_FIELD_NUMBER: ClassVar[int]
    RESULT_FIELD_NUMBER: ClassVar[int]
    metadata: Metadata
    result: GetFeaturesResult
    def __init__(self, result: Optional[Union[GetFeaturesResult, Mapping]] = ..., metadata: Optional[Union[Metadata, Mapping]] = ...) -> None: ...

class GetFeaturesResult(_message.Message):
    __slots__ = ["features", "join_keys"]
    FEATURES_FIELD_NUMBER: ClassVar[int]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    features: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Value]
    join_keys: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Value]
    def __init__(self, features: Optional[Iterable[Union[_struct_pb2.Value, Mapping]]] = ..., join_keys: Optional[Iterable[Union[_struct_pb2.Value, Mapping]]] = ...) -> None: ...

class Metadata(_message.Message):
    __slots__ = ["features", "join_keys", "partial_results", "slo_info"]
    class Item(_message.Message):
        __slots__ = ["data_type", "description", "effective_time", "name", "status", "tags", "type"]
        class TagsEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: ClassVar[int]
            VALUE_FIELD_NUMBER: ClassVar[int]
            key: str
            value: str
            def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
        DATA_TYPE_FIELD_NUMBER: ClassVar[int]
        DESCRIPTION_FIELD_NUMBER: ClassVar[int]
        EFFECTIVE_TIME_FIELD_NUMBER: ClassVar[int]
        NAME_FIELD_NUMBER: ClassVar[int]
        STATUS_FIELD_NUMBER: ClassVar[int]
        TAGS_FIELD_NUMBER: ClassVar[int]
        TYPE_FIELD_NUMBER: ClassVar[int]
        data_type: FeatureServerComplexDataType
        description: str
        effective_time: _timestamp_pb2.Timestamp
        name: str
        status: Status
        tags: _containers.ScalarMap[str, str]
        type: FeatureServerDataType
        def __init__(self, name: Optional[str] = ..., effective_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., type: Optional[Union[FeatureServerDataType, str]] = ..., data_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ..., status: Optional[Union[Status, str]] = ..., description: Optional[str] = ..., tags: Optional[Mapping[str, str]] = ...) -> None: ...
    FEATURES_FIELD_NUMBER: ClassVar[int]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    PARTIAL_RESULTS_FIELD_NUMBER: ClassVar[int]
    SLO_INFO_FIELD_NUMBER: ClassVar[int]
    features: _containers.RepeatedCompositeFieldContainer[Metadata.Item]
    join_keys: _containers.RepeatedCompositeFieldContainer[Metadata.Item]
    partial_results: bool
    slo_info: SloInfo
    def __init__(self, features: Optional[Iterable[Union[Metadata.Item, Mapping]]] = ..., join_keys: Optional[Iterable[Union[Metadata.Item, Mapping]]] = ..., slo_info: Optional[Union[SloInfo, Mapping]] = ..., partial_results: bool = ...) -> None: ...

class NameAndType(_message.Message):
    __slots__ = ["data_type", "name", "type"]
    DATA_TYPE_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    data_type: FeatureServerComplexDataType
    name: str
    type: FeatureServerDataType
    def __init__(self, name: Optional[str] = ..., data_type: Optional[Union[FeatureServerComplexDataType, Mapping]] = ..., type: Optional[Union[FeatureServerDataType, str]] = ...) -> None: ...

class QueryFeaturesRequest(_message.Message):
    __slots__ = ["params"]
    PARAMS_FIELD_NUMBER: ClassVar[int]
    params: _feature_service_request__client_pb2.GetFeaturesParameters
    def __init__(self, params: Optional[Union[_feature_service_request__client_pb2.GetFeaturesParameters, Mapping]] = ...) -> None: ...

class QueryFeaturesResponse(_message.Message):
    __slots__ = ["metadata", "results"]
    METADATA_FIELD_NUMBER: ClassVar[int]
    RESULTS_FIELD_NUMBER: ClassVar[int]
    metadata: Metadata
    results: _containers.RepeatedCompositeFieldContainer[GetFeaturesResult]
    def __init__(self, results: Optional[Iterable[Union[GetFeaturesResult, Mapping]]] = ..., metadata: Optional[Union[Metadata, Mapping]] = ...) -> None: ...

class ServiceMetadataRequest(_message.Message):
    __slots__ = ["params"]
    PARAMS_FIELD_NUMBER: ClassVar[int]
    params: FeatureServiceLocator
    def __init__(self, params: Optional[Union[FeatureServiceLocator, Mapping]] = ...) -> None: ...

class ServiceMetadataResponse(_message.Message):
    __slots__ = ["feature_service_type", "feature_values", "input_join_keys", "input_request_context_keys", "output_join_keys"]
    FEATURE_SERVICE_TYPE_FIELD_NUMBER: ClassVar[int]
    FEATURE_VALUES_FIELD_NUMBER: ClassVar[int]
    INPUT_JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    INPUT_REQUEST_CONTEXT_KEYS_FIELD_NUMBER: ClassVar[int]
    OUTPUT_JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    feature_service_type: FeatureServiceType
    feature_values: _containers.RepeatedCompositeFieldContainer[NameAndType]
    input_join_keys: _containers.RepeatedCompositeFieldContainer[NameAndType]
    input_request_context_keys: _containers.RepeatedCompositeFieldContainer[NameAndType]
    output_join_keys: _containers.RepeatedCompositeFieldContainer[NameAndType]
    def __init__(self, feature_service_type: Optional[Union[FeatureServiceType, str]] = ..., input_join_keys: Optional[Iterable[Union[NameAndType, Mapping]]] = ..., input_request_context_keys: Optional[Iterable[Union[NameAndType, Mapping]]] = ..., output_join_keys: Optional[Iterable[Union[NameAndType, Mapping]]] = ..., feature_values: Optional[Iterable[Union[NameAndType, Mapping]]] = ...) -> None: ...

class SloInfo(_message.Message):
    __slots__ = ["dynamodb_response_size_bytes", "server_time_seconds", "slo_eligible", "slo_ineligibility_reasons", "slo_server_time_seconds", "store_max_latency", "store_response_size_bytes", "store_time_seconds"]
    DYNAMODB_RESPONSE_SIZE_BYTES_FIELD_NUMBER: ClassVar[int]
    SERVER_TIME_SECONDS_FIELD_NUMBER: ClassVar[int]
    SLO_ELIGIBLE_FIELD_NUMBER: ClassVar[int]
    SLO_INELIGIBILITY_REASONS_FIELD_NUMBER: ClassVar[int]
    SLO_SERVER_TIME_SECONDS_FIELD_NUMBER: ClassVar[int]
    STORE_MAX_LATENCY_FIELD_NUMBER: ClassVar[int]
    STORE_RESPONSE_SIZE_BYTES_FIELD_NUMBER: ClassVar[int]
    STORE_TIME_SECONDS_FIELD_NUMBER: ClassVar[int]
    dynamodb_response_size_bytes: int
    server_time_seconds: float
    slo_eligible: bool
    slo_ineligibility_reasons: _containers.RepeatedScalarFieldContainer[SloIneligibilityReason]
    slo_server_time_seconds: float
    store_max_latency: float
    store_response_size_bytes: int
    store_time_seconds: float
    def __init__(self, slo_eligible: bool = ..., slo_server_time_seconds: Optional[float] = ..., slo_ineligibility_reasons: Optional[Iterable[Union[SloIneligibilityReason, str]]] = ..., dynamodb_response_size_bytes: Optional[int] = ..., server_time_seconds: Optional[float] = ..., store_time_seconds: Optional[float] = ..., store_max_latency: Optional[float] = ..., store_response_size_bytes: Optional[int] = ...) -> None: ...

class FeatureServiceType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class FeatureServerDataType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class Status(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class SloIneligibilityReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.data import entity__client_pb2 as _entity__client_pb2
from tecton_proto.data import feature_service__client_pb2 as _feature_service__client_pb2
from tecton_proto.data import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.data import server_group__client_pb2 as _server_group__client_pb2
from tecton_proto.data import transformation__client_pb2 as _transformation__client_pb2
from tecton_proto.data import virtual_data_source__client_pb2 as _virtual_data_source__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class Fco(_message.Message):
    __slots__ = ["entity", "feature_service", "feature_view", "server_group", "transformation", "virtual_data_source"]
    ENTITY_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: ClassVar[int]
    SERVER_GROUP_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATION_FIELD_NUMBER: ClassVar[int]
    VIRTUAL_DATA_SOURCE_FIELD_NUMBER: ClassVar[int]
    entity: _entity__client_pb2.Entity
    feature_service: _feature_service__client_pb2.FeatureService
    feature_view: _feature_view__client_pb2.FeatureView
    server_group: _server_group__client_pb2.ServerGroup
    transformation: _transformation__client_pb2.Transformation
    virtual_data_source: _virtual_data_source__client_pb2.VirtualDataSource
    def __init__(self, virtual_data_source: Optional[Union[_virtual_data_source__client_pb2.VirtualDataSource, Mapping]] = ..., entity: Optional[Union[_entity__client_pb2.Entity, Mapping]] = ..., feature_view: Optional[Union[_feature_view__client_pb2.FeatureView, Mapping]] = ..., feature_service: Optional[Union[_feature_service__client_pb2.FeatureService, Mapping]] = ..., transformation: Optional[Union[_transformation__client_pb2.Transformation, Mapping]] = ..., server_group: Optional[Union[_server_group__client_pb2.ServerGroup, Mapping]] = ...) -> None: ...

class FcoContainer(_message.Message):
    __slots__ = ["fcos", "root_ids", "workspace", "workspace_state_id"]
    FCOS_FIELD_NUMBER: ClassVar[int]
    ROOT_IDS_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_STATE_ID_FIELD_NUMBER: ClassVar[int]
    fcos: _containers.RepeatedCompositeFieldContainer[Fco]
    root_ids: _containers.RepeatedCompositeFieldContainer[_id__client_pb2.Id]
    workspace: str
    workspace_state_id: _id__client_pb2.Id
    def __init__(self, root_ids: Optional[Iterable[Union[_id__client_pb2.Id, Mapping]]] = ..., fcos: Optional[Iterable[Union[Fco, Mapping]]] = ..., workspace: Optional[str] = ..., workspace_state_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class User(_message.Message):
    __slots__ = ["created_at", "first_name", "is_admin", "last_login", "last_name", "login_email", "okta_id", "okta_status"]
    CREATED_AT_FIELD_NUMBER: ClassVar[int]
    FIRST_NAME_FIELD_NUMBER: ClassVar[int]
    IS_ADMIN_FIELD_NUMBER: ClassVar[int]
    LAST_LOGIN_FIELD_NUMBER: ClassVar[int]
    LAST_NAME_FIELD_NUMBER: ClassVar[int]
    LOGIN_EMAIL_FIELD_NUMBER: ClassVar[int]
    OKTA_ID_FIELD_NUMBER: ClassVar[int]
    OKTA_STATUS_FIELD_NUMBER: ClassVar[int]
    created_at: _timestamp_pb2.Timestamp
    first_name: str
    is_admin: bool
    last_login: _timestamp_pb2.Timestamp
    last_name: str
    login_email: str
    okta_id: str
    okta_status: str
    def __init__(self, okta_id: Optional[str] = ..., first_name: Optional[str] = ..., last_name: Optional[str] = ..., login_email: Optional[str] = ..., okta_status: Optional[str] = ..., created_at: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., is_admin: bool = ..., last_login: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Optional

DESCRIPTOR: _descriptor.FileDescriptor

class HiveTableDataSource(_message.Message):
    __slots__ = ["database", "table"]
    DATABASE_FIELD_NUMBER: ClassVar[int]
    TABLE_FIELD_NUMBER: ClassVar[int]
    database: str
    table: str
    def __init__(self, table: Optional[str] = ..., database: Optional[str] = ...) -> None: ...

class ListHiveResult(_message.Message):
    __slots__ = ["names"]
    NAMES_FIELD_NUMBER: ClassVar[int]
    names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, names: Optional[Iterable[str]] = ...) -> None: ...



import contextlib
import logging
import os
from datetime import datetime
from typing import Optional

import pyarrow.fs
import ray
from google.protobuf.json_format import MessageToJson

from tecton_core import conf
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.query.query_tree_executor import QueryTreeExecutor
from tecton_core.query.snowflake.compute import SnowflakeCompute
from tecton_core.secrets import SecretResolver
from tecton_core.snowflake_context import SnowflakeContext
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.common.task_params import redact_sensitive_fields_from_params
from tecton_materialization.ray.batch_materialization import run_batch_materialization
from tecton_materialization.ray.dataset_generation import run_dataset_generation
from tecton_materialization.ray.delta import DeltaWriter
from tecton_materialization.ray.feature_export import get_feature_export_qt
from tecton_materialization.ray.feature_export import get_feature_export_store_params
from tecton_materialization.ray.ingest_materialization import ingest_pushed_df
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.job_status import MonitoringContextProvider
from tecton_materialization.ray.materialization_utils import get_delta_writer
from tecton_materialization.ray.materialization_utils import get_delta_writer_for_fd
from tecton_materialization.ray.materialization_utils import run_online_store_copier
from tecton_materialization.secrets import MDSSecretResolver
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.materialization.params__client_pb2 import SecretMaterializationTaskParams
from tecton_proto.materialization.params__client_pb2 import SecretServiceParams
from tecton_proto.offlinestore.delta import metadata__client_pb2 as metadata_pb2
from tecton_proto.online_store_writer.copier__client_pb2 import DeletionRequest
from tecton_proto.online_store_writer.copier__client_pb2 import OnlineStoreCopierRequest


logger = logging.getLogger(__name__)

FEATURE_EXPORT_TASK_TYPE = "feature_export"

_DIALECT_TO_STAGE_TYPE = {
    Dialect.PANDAS: TectonManagedStage.PYTHON,
    Dialect.DUCKDB: TectonManagedStage.PYTHON,
    Dialect.SNOWFLAKE: TectonManagedStage.SNOWFLAKE,
}

_DIALECT_TO_UI_STRING = {
    Dialect.PANDAS: "Python",
    Dialect.DUCKDB: "Python",
    Dialect.SNOWFLAKE: "Snowflake",
}


def _get_compute(dialect: Dialect, qt: NodeRef, secret_resolver: Optional[SecretResolver]) -> SQLCompute:
    if dialect == Dialect.SNOWFLAKE:
        if SnowflakeContext.is_initialized():
            return SnowflakeCompute.for_connection(SnowflakeContext.get_instance().get_connection())
        else:
            return SnowflakeCompute.for_query_tree(qt, secret_resolver)
    return SQLCompute.for_dialect(dialect)


def _delete_from_online_store(
    materialization_task_params: MaterializationTaskParams, job_status_client: JobStatusClient
) -> None:
    online_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.ONLINE_STORE,
        "Unload features to online store",
    )
    with online_stage_monitor() as progress_callback:
        if materialization_task_params.deletion_task_info.deletion_parameters.HasField("online_join_keys_path"):
            deletion_request = DeletionRequest(
                online_join_keys_path=materialization_task_params.deletion_task_info.deletion_parameters.online_join_keys_path,
            )
        else:
            deletion_request = DeletionRequest(
                online_join_keys_full_path=materialization_task_params.deletion_task_info.deletion_parameters.online_join_keys_full_path,
            )
        request = OnlineStoreCopierRequest(
            online_store_writer_configuration=materialization_task_params.online_store_writer_config,
            feature_view=materialization_task_params.feature_view,
            deletion_request=deletion_request,
        )
        run_online_store_copier(request)
    progress_callback(1.0)


def _delete_from_offline_store(params: MaterializationTaskParams, job_status_client: JobStatusClient):
    offline_uri = params.deletion_task_info.deletion_parameters.offline_join_keys_path
    fs, path = pyarrow.fs.FileSystem.from_uri(offline_uri)
    keys_table = pyarrow.dataset.dataset(source=path, filesystem=fs).to_table()
    offline_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE, "Delete keys from offline store"
    )
    with offline_stage_monitor():
        delta_writer = get_delta_writer_for_fd(params)
        delta_writer.delete_keys(keys_table)
        delta_writer.commit()


def _feature_export(
    fd: FeatureDefinitionWrapper,
    task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    export_params = task_params.feature_export_info.feature_export_parameters
    start_time = export_params.feature_start_time.ToDatetime()
    end_time = export_params.feature_end_time.ToDatetime()
    table_uri = export_params.export_store_path
    store_params = get_feature_export_store_params(fd)
    delta_writer = get_delta_writer(task_params, store_params=store_params, table_uri=table_uri, join_keys=fd.join_keys)
    delta_write_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE,
        "Write full features to offline store.",
    )

    _run_export(fd, start_time, end_time, executor, delta_writer, delta_write_monitor)


def _run_export(
    fd: FeatureDefinitionWrapper,
    start_time: datetime,
    end_time: datetime,
    qt_executor: QueryTreeExecutor,
    delta_writer: DeltaWriter,
    write_monitor: MonitoringContextProvider,
):
    qt, interval = get_feature_export_qt(fd, start_time, end_time)
    feature_data = qt_executor.exec_qt(qt).result_table

    with write_monitor():
        # TODO (TEC-18865): add support for is_overwrite flag for feature_export jobs
        transaction_metadata = metadata_pb2.TectonDeltaMetadata()
        transaction_metadata.feature_start_time.FromDatetime(interval.start)

        @delta_writer.transaction(transaction_metadata)
        def txn():
            delta_writer.delete_time_range(interval)
            delta_writer.write(feature_data)

        txn()


@contextlib.contextmanager
def _ray():
    logging.warning(f"Initializing Ray from classpath: {os.environ['CLASSPATH']}")
    ray.init(
        job_config=ray.job_config.JobConfig(
            code_search_path=os.environ["CLASSPATH"].split(":"),
            # Increasing max heap size mostly for Online Store Copier.
            # Our target Parquet file size is around 1GB (although, not guaranteed, see delta.py),
            # so to download and decode such file 2-4GB might be needed.
            # Multiplying it by 2x factor for some buffer.
            jvm_options=["-Xmx8G"],
        ),
        include_dashboard=False,
    )
    try:
        yield
    finally:
        ray.shutdown()


def run_materialization(
    materialization_task_params: MaterializationTaskParams,
    secret_materialization_task_params: SecretMaterializationTaskParams,
) -> None:
    json_params = MessageToJson(materialization_task_params)
    redacted_json_params = redact_sensitive_fields_from_params(json_params)
    logger.info(f"Starting materialization with params: {redacted_json_params}")

    conf.set("DUCKDB_DEBUG", "true")
    conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "rift")
    conf.set("TECTON_RUNTIME_MODE", "MATERIALIZATION")
    job_status_client = JobStatusClient(JobMetadataClient.for_params(materialization_task_params))
    job_status_client.set_overall_state(TectonManagedStage.State.RUNNING)
    try:
        run_ray_job(materialization_task_params, secret_materialization_task_params, job_status_client)
        job_status_client.set_overall_state(TectonManagedStage.State.SUCCESS)
    except Exception:
        job_status_client.set_current_stage_failed(TectonManagedStage.ErrorType.UNEXPECTED_ERROR)
        job_status_client.set_overall_state(TectonManagedStage.State.ERROR)
        raise


def _run_maintenance(materialization_task_params: MaterializationTaskParams, fd: FeatureDefinitionWrapper) -> None:
    delta_writer = get_delta_writer_for_fd(materialization_task_params)
    maintenance_params = materialization_task_params.delta_maintenance_task_info.delta_maintenance_parameters

    if maintenance_params.execute_sorting:
        delta_writer.run_z_order_optimization()
        delta_writer.commit()
    if maintenance_params.vacuum:
        # vacuuming is not implemented
        pass


def run_ray_job(
    materialization_task_params: MaterializationTaskParams,
    secret_materialization_task_params: SecretMaterializationTaskParams,
    job_status_client: JobStatusClient,
) -> None:
    secret_resolver = _get_secret_resolver(secret_materialization_task_params.secret_service_params)
    task_type = _get_task_type(materialization_task_params)
    executor = QueryTreeExecutor(monitor=job_status_client, secret_resolver=secret_resolver)

    if task_type == "deletion_task":
        _delete_from_offline_store(materialization_task_params, job_status_client)
        _delete_from_online_store(materialization_task_params, job_status_client)
    elif task_type == "delta_maintenance_task":
        # TODO (TEC-19141): delta-rs maintenance commits are incompatible with delta standalone library
        fd = feature_definition_from_task_params(materialization_task_params)

        _run_maintenance(materialization_task_params, fd)
    elif task_type == "ingest_task":
        assert materialization_task_params.feature_view.schemas.HasField("materialization_schema"), "missing schema"
        fd = feature_definition_from_task_params(materialization_task_params)
        ingest_pushed_df(materialization_task_params, fd, job_status_client)
    elif task_type == FEATURE_EXPORT_TASK_TYPE:
        fd = feature_definition_from_task_params(materialization_task_params)
        _feature_export(fd, materialization_task_params, job_status_client, executor)
    elif task_type == "batch_task":
        assert materialization_task_params.feature_view.schemas.HasField("materialization_schema"), "missing schema"

        run_batch_materialization(materialization_task_params, job_status_client, executor)
    elif task_type == "dataset_generation_task":
        run_dataset_generation(materialization_task_params, job_status_client, executor)
    else:
        msg = f"Task type {task_type} is not supported by Ray materialization job"
        raise ValueError(msg)


def _get_task_type(materialization_task_params: MaterializationTaskParams) -> str:
    return materialization_task_params.WhichOneof("task_info")[:-5]  # removesuffix("_info")


def _get_secret_resolver(
    secret_service_params: SecretServiceParams,
) -> Optional[SecretResolver]:
    if not secret_service_params.secrets_api_service_url:
        return None

    assert secret_service_params.secret_access_api_key, "Secret access key is required when using secret service"
    return MDSSecretResolver(
        secret_service_params.secrets_api_service_url,
        secret_service_params.secret_access_api_key,
    )



from types import MappingProxyType
from typing import Mapping
from typing import Union

from typeguard import typechecked

from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import server_group__client_pb2 as server_group__arg_pb2
from tecton_proto.common import framework_version__client_pb2 as framework_version_pb2
from tecton_proto.common import scaling_config__client_pb2 as scaling_config_pb2
from tecton_proto.data import server_group__client_pb2 as server_group__data_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = [
    "FeatureServerGroupSpec",
    "TransformServerGroupSpec",
]


@utils.frozen_strict
class FeatureServerGroupSpec(tecton_object_spec.TectonObjectSpec):
    """Base class for feature server group specs."""

    scaling_config: Union[scaling_config_pb2.AutoscalingConfig, scaling_config_pb2.ProvisionedScalingConfig]
    options: Mapping[str, str]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: server_group__data_pb2.ServerGroup) -> "FeatureServerGroupSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.server_group_id, proto.fco_metadata
            ),
            scaling_config=get_scaling_config(proto),
            validation_args=validator_pb2.FcoValidationArgs(server_group=proto.validation_args),
            options=MappingProxyType(proto.options),
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: server_group__arg_pb2.ServerGroupArgs) -> "FeatureServerGroupSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(
                proto.server_group_id, proto.info, framework_version_pb2.FrameworkVersion.FWV6
            ),
            scaling_config=get_scaling_config(proto),
            validation_args=None,
            options=MappingProxyType(proto.options),
        )


@utils.frozen_strict
class TransformServerGroupSpec(tecton_object_spec.TectonObjectSpec):
    """Base class for feature server group specs."""

    scaling_config: Union[scaling_config_pb2.AutoscalingConfig, scaling_config_pb2.ProvisionedScalingConfig]
    environment: str
    options: Mapping[str, str]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: server_group__data_pb2.ServerGroup) -> "TransformServerGroupSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.server_group_id, proto.fco_metadata
            ),
            scaling_config=get_scaling_config(proto),
            environment=proto.transform_server_group.environment_name,
            validation_args=validator_pb2.FcoValidationArgs(server_group=proto.validation_args),
            options=MappingProxyType(proto.options),
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: server_group__arg_pb2.ServerGroupArgs) -> "TransformServerGroupSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(
                proto.server_group_id, proto.info, framework_version_pb2.FrameworkVersion.FWV6
            ),
            scaling_config=get_scaling_config(proto),
            environment=proto.transform_server_group_args.environment,
            validation_args=None,
            options=MappingProxyType(proto.options),
        )


def get_scaling_config(
    proto: Union[server_group__arg_pb2.ServerGroupArgs, server_group__data_pb2.ServerGroup],
) -> Union[scaling_config_pb2.AutoscalingConfig, scaling_config_pb2.ProvisionedScalingConfig]:
    if proto.HasField("autoscaling_config"):
        return proto.autoscaling_config
    elif proto.HasField("provisioned_scaling_config"):
        return proto.provisioned_scaling_config
    else:
        msg = f"Unexpected scaling config type: {proto}"
        raise ValueError(msg)



from typing import Tuple

from tecton._internals import errors
from tecton.framework import base_tecton_object
from tecton_proto.args import transformation__client_pb2 as transformation__args_proto


SPARK_SQL_MODE = "spark_sql"
PYSPARK_MODE = "pyspark"
SNOWFLAKE_SQL_MODE = "snowflake_sql"
SNOWPARK_MODE = "snowpark"
PANDAS_MODE = "pandas"
PYTHON_MODE = "python"
BIGQUERY_SQL_MODE = "bigquery_sql"

mode_str_to_proto_enum = {
    SPARK_SQL_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_SPARK_SQL,
    PYSPARK_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_PYSPARK,
    SNOWFLAKE_SQL_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_SNOWFLAKE_SQL,
    SNOWPARK_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_SNOWPARK,
    PANDAS_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_PANDAS,
    PYTHON_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_PYTHON,
    BIGQUERY_SQL_MODE: transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_BIGQUERY_SQL,
}


def short_tecton_objects_repr(tecton_objects: Tuple[base_tecton_object.BaseTectonObject]) -> str:
    """Returns a shortened printable representation for a tuple of Tecton objects. Used for printing summaries."""
    short_strings = tuple(short_tecton_object_repr(obj) for obj in tecton_objects)
    return repr(short_strings)


def short_tecton_object_repr(tecton_object: base_tecton_object.BaseTectonObject) -> str:
    """Returns a shortened printable representation for a Tecton object. Used for printing summaries."""
    return f"{type(tecton_object).__name__}('{tecton_object.info.name}')"


def get_transformation_mode_name(mode_proto_enum: mode_str_to_proto_enum) -> str:
    mode_proto_enum_to_str = dict(map(reversed, mode_str_to_proto_enum.items()))
    return mode_proto_enum_to_str.get(mode_proto_enum)


def get_transformation_mode_enum(mode: str, name: str) -> transformation__args_proto.TransformationMode.ValueType:
    """Returns the TransformationMode type from string"""
    mode_enum = mode_str_to_proto_enum.get(mode)
    if mode_enum is None:
        raise errors.InvalidTransformationMode(
            name,
            mode,
            [SPARK_SQL_MODE, PYSPARK_MODE, SNOWFLAKE_SQL_MODE, SNOWPARK_MODE, PANDAS_MODE, PYTHON_MODE],
        )
    else:
        return mode_enum



import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import List
from typing import Optional

import click
from google.protobuf.json_format import MessageToJson
from texttable import Texttable

from tecton import tecton_context
from tecton._internals import metadata_service
from tecton.cli import printer
from tecton.cli.cli_utils import display_table
from tecton.cli.cli_utils import timestamp_to_string
from tecton.cli.command import TectonGroup
from tecton_proto.common.server_group_status__client_pb2 import ServerGroupStatus
from tecton_proto.common.server_group_type__client_pb2 import ServerGroupType
from tecton_proto.servergroupservice.server_group_service__client_pb2 import GetServerGroupRequest
from tecton_proto.servergroupservice.server_group_service__client_pb2 import ListServerGroupsRequest
from tecton_proto.servergroupservice.server_group_service__client_pb2 import ServerGroupInfo


logger = logging.getLogger(__name__)

ERROR_MESSAGE_PREFIX = "⛔ ERROR: "
CHECK_MARK = "✅"
ERROR_SIGN = "⛔"
INFO_SIGN = "💡"


@dataclass
class ServerGroupIdentifier:
    id: Optional[str]
    name: Optional[str]

    def __post_init__(self):
        if not self.id and not self.name:
            printer.safe_print(
                f"{ERROR_MESSAGE_PREFIX} At least one of `id` or `name` must be provided", file=sys.stderr
            )
            sys.exit(1)

    def __str__(self):
        if self.id:
            return f"id: {self.id}"
        elif self.name:
            return f"name: {self.name}"
        else:
            return "No name or id set"

    def __eq__(self, identifier):
        if isinstance(identifier, ServerGroupIdentifier):
            if self.id:
                return self.id == identifier.id
            elif self.name:
                return self.name == identifier.name
        return False


@click.command("server-group", cls=TectonGroup)
def server_group():
    """Manage Server Groups for Online Serving and Realtime Execution."""


@server_group.command("list")
def list():
    """List all available Server Groups in workspace"""
    workspace = tecton_context.get_current_workspace()
    server_groups = _list_server_groups(workspace)
    _display_server_groups(server_groups)


@server_group.command("describe")
@click.option("-n", "--name", help="Name of the Server Group", required=True, type=str)
@click.option(
    "-o",
    "--output-file",
    help="JSON Output file to write Server Group Information to. If not specified, the Server Group information will be printed to stdout",
    required=False,
    type=click.Path(exists=False),
)
def describe(name: str, output_file: Optional[str] = None):
    """Describe a Server Group"""
    workspace = tecton_context.get_current_workspace()
    group = _get_server_group(workspace, name)
    if output_file:
        output_path = Path(output_file)
        _write_to_json(group, output_path)
    else:
        _display_server_group_info(group)


def _list_server_groups(workspace: str, type: ServerGroupType = None):
    try:
        req = ListServerGroupsRequest(workspace=workspace)
        if type:
            req.type = type
        response = metadata_service.instance().ListServerGroups(req)
        return response.server_groups

    except Exception as e:
        printer.safe_print(f"{ERROR_MESSAGE_PREFIX} Failed to fetch Server Groups: {e}", file=sys.stderr)
        sys.exit(1)


def _display_server_groups(server_groups: List[ServerGroupInfo], type_filter: ServerGroupType = None):
    headings = [
        "Id",
        "Name",
        "Type",
        "Status",
        "Environment",
        "Description",
        "Created At",
        "Owner",
        "Last Modified By",
    ]
    if type_filter and type_filter == ServerGroupType.SERVER_GROUP_TYPE_TRANSFORM_SERVER_GROUP:
        headings.append("Environment")
    display_table(
        headings,
        [
            (
                group.server_group_id,
                group.name,
                ServerGroupType.Name(group.type).split("SERVER_GROUP_TYPE_")[-1],
                ServerGroupStatus.Name(group.status).split("SERVER_GROUP_STATUS_")[-1],
                group.environment or "None"
                if group.type is ServerGroupType.SERVER_GROUP_TYPE_TRANSFORM_SERVER_GROUP
                else "N/A",
                group.description or "None",
                timestamp_to_string(group.created_at),
                group.owner,
                group.last_modified_by,
            )
            for group in server_groups
        ],
    )


def _get_server_group(workspace: str, server_group_name: str) -> ServerGroupInfo:
    try:
        req = GetServerGroupRequest(workspace=workspace, server_group_name=server_group_name)
        response = metadata_service.instance().GetServerGroup(req)
        return response.server_group

    except Exception as e:
        printer.safe_print(f"{ERROR_MESSAGE_PREFIX} Failed to fetch Server Groups: {e}", file=sys.stderr)
        sys.exit(1)


def _display_server_group_info(group: ServerGroupInfo):
    def _display_table(headings, rows):
        table = Texttable()
        table.header(headings)
        table.add_rows(rows, header=False)
        print(table.draw())
        print()

    def print_section_header(title):
        print("\n" + "=" * len(title))
        print(title.upper())
        print("=" * len(title))

    # Server Group Details
    print_section_header("Server Group Details")
    details = [
        ("ID", group.server_group_id),
        ("Name", group.name),
        ("Type", ServerGroupType.Name(group.type).split("SERVER_GROUP_TYPE_")[-1]),
        ("Status", ServerGroupStatus.Name(group.status).split("SERVER_GROUP_STATUS_")[-1]),
        ("Status Details", group.status_details or "N/A"),
        ("Created At", timestamp_to_string(group.created_at)),
        ("Description", group.description or "N/A"),
        ("Owner", group.owner or "N/A"),
        ("Last Modified By", group.last_modified_by),
    ]
    if group.type == ServerGroupType.SERVER_GROUP_TYPE_TRANSFORM_SERVER_GROUP:
        details.append(("Environment", group.environment))

    _display_table(["Field", "Value"], details)

    # Scaling Configurations
    desired_autoscaling_enabled = group.desired_config.autoscaling_enabled
    current_autoscaling_enabled = group.current_config.autoscaling_enabled

    scaling_config = [
        (
            "Autoscaling Enabled",
            bool_to_string(desired_autoscaling_enabled),
            bool_to_string(current_autoscaling_enabled),
        ),
        (
            "Min Nodes\n(Autoscaling Config)",
            "N/A" if not desired_autoscaling_enabled else group.desired_config.min_nodes,
            "N/A" if not current_autoscaling_enabled else group.current_config.min_nodes,
        ),
        (
            "Max Nodes\n(Autoscaling Config)",
            "N/A" if not desired_autoscaling_enabled else group.desired_config.max_nodes,
            "N/A" if not current_autoscaling_enabled else group.current_config.max_nodes,
        ),
        (
            "Desired Nodes\n(Provisioned Config)",
            "N/A" if desired_autoscaling_enabled else group.desired_config.desired_nodes,
            "N/A" if current_autoscaling_enabled else group.current_config.desired_nodes,
        ),
        (
            "Last Updated At",
            timestamp_to_string(group.desired_config.last_updated_at),
            timestamp_to_string(group.current_config.last_updated_at),
        ),
    ]

    print_section_header("Scaling Configurations")
    _display_table(["Field", "Target Configuration", "Active Configuration"], scaling_config)


def bool_to_string(value):
    return "True" if bool(value) else "False"


def _write_to_json(server_group_info: ServerGroupInfo, output_path: Path):
    server_group_json = MessageToJson(server_group_info)
    with output_path.open("w") as f:
        f.write(server_group_json)
